.container-fluid {
    width: 80%;
    padding-right: 15px;
    padding-left: 15px;
    margin-right: auto;
    margin-left: auto;
}

.right {
  float: right;
  width: 300px;
  border: 3px solid #73AD21;
  padding: 10px;
}

.card {
    border: 0;
    margin-bottom: 30px;
    margin-top: 10px;
    border-radius: 6px;
    color: #333;
    background: #fff;
    width: 100%;
    box-shadow: 0 2px 2px 0 rgb(0 0 0 / 14%), 0 3px 1px -2px rgb(0 0 0 / 20%), 0 1px 5px 0 rgb(0 0 0 / 12%);
}

.main-panel>.content {
    margin-top: 0px;
    padding: 30px 15px;
    min-height: calc(100vh - 123px);
}

a.disabled, fieldset:disabled a {
    pointer-events: none;
    background-color: #D8D8D8;
}


.list-group-item {
    position: relative;
    display: block;
    padding: 0.25rem 1.25rem;
    margin-bottom: 0;
    background-color: inherit;
    border: 0 solid rgba(0,0,0,.125);
}

#Cluster\ Select {
    height: 150px
}

#overlay { width:100px;
    height: 100px;
    position: fixed;
    top: 50%;
    left: 50%;
    z-index: -1;
}

#loading { width:100px;
    height: 200px;
    position: fixed;
    top: 50%;
    left: 50%;
}

.btn, .btn.btn-default {
    color: #fff;
    background-color: #9124a3;
    border-color: #9c27b0;
    box-shadow: 0 2px 2px 0 hsl(0deg 0% 60% / 14%), 0 3px 1px -2px hsl(0deg 0% 60% / 20%), 0 1px 5px 0 hsl(0deg 0% 60% / 12%);
}

.logging_window{
    display: block;
    padding: 9.5px;
    font-size: 13px;
    line-height: 1.42857143;
    color: #333;
    word-break: break-all;
    word-wrap: break-word;
    background-color: #f5f5f5;
    border: 1px solid #ccc;
    border-radius: 4px;
    width: 50%;
    margin: auto;
}

.sidebar {
    width: 285px;
}

.sidebar .sidebar-wrapper {
    width: 285px;
}
