{% extends 'base.html' %}
{% block content %}

<div >
	<a class="btn" href="/choose_cluster">Choose cluster result to run PSTS</a>
</div>
<br>

<div class="card">
	<div class="card-header card-header-primary">
		<h4 class="card-title">Pseudo-time-space analysis</h4>
	</div>
	<div class="card-body">

	{# Below indicates whether clustering already completed #}
	{% include "flash_header.html" %}

	{# Below is the form. #}
    {% with superForm=psts_form %}
    {% include "superform.html" %}
    {% endwith %}
	</div>
</div>
{% endblock %}

{% block status %}
{% include "progress.html" %}
{% endblock %}
